# Author: Franziska Horn <cod3licious@gmail.com>
# License: MIT

from __future__ import annotations

import logging
import sys
from collections import defaultdict

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.preprocessing import PowerTransformer, StandardScaler
from sklearn.utils.validation import check_array, check_is_fitted

logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s", level=logging.INFO)


def _check_features(df: pd.DataFrame, corrthr: float = 0.995, verbose: int = 0) -> list:
    """
    Identify features with zeros variance or a correlation of (almost) 1 to other features, i.e., useless features.

    Inputs:
        - df: pd.DataFrame with all features
        - corrthr: threshold for correlations: if a feature has a higher pearson correlation to another feature it's
                   considered as redundant and ignored (float; default: 0.995)
        - verbose: verbosity level (int; default: 0)

    Returns:
        - list of column names representing 'ok' features (numeric, non-zero variance, not redundant)
    """
    # make sure all data is numeric
    df = df.select_dtypes(include=np.number)
    useless_cols = set()
    # 1. identify features with zero variance
    for c, v in df.var().items():
        if pd.isna(v) or v <= sys.float_info.epsilon:
            useless_cols.add(c)
    # 2. identify redundant features (that have a correlation of ~1 with other features)
    correlated_cols = defaultdict(set)
    corrmat = pd.DataFrame(np.abs(np.corrcoef(df.values, rowvar=False)), columns=df.columns, index=df.columns)
    np.fill_diagonal(corrmat.values, 0)
    for c, v in corrmat.unstack().sort_values(ascending=False).items():  # noqa PD010
        if v < corrthr:
            break
        if (c[0] != c[1]) and (c[0] not in useless_cols):
            correlated_cols[c[0]].add(c[1])
    # keep the columns that eliminate the most correlated columns
    for c in sorted(correlated_cols, key=lambda x: len(correlated_cols[x]), reverse=True):
        # the first variable that is correlated with others adds its correlated variables to the set of useless cols
        # since we check if a variable is in useless_cols, the correlated variables can't add the original variable
        if c not in useless_cols:
            useless_cols.update(correlated_cols[c])
    # return list of columns that should be kept
    if verbose:
        logging.info(f"[AutoFeatLight] {len(useless_cols)} columns identified as useless:")
        logging.info(sorted(useless_cols))
    return [c for c in df.columns if c not in useless_cols]


def _compute_additional_features(
    X: np.ndarray,
    feature_names: list | None = None,
    compute_ratio: bool = True,
    compute_product: bool = True,
    verbose: int = 0,
) -> tuple[np.ndarray | None, list]:
    """
    Compute additional non-linear features from the original features (ratio or product of two features).

    Inputs:
        - X: np.array with data (n_datapoints x n_features)
        - feature_names: optional list of column names to identify the features in X
        - compute_ratio: bool (default: True), whether to compute ratios of features
        - compute_product: bool (default: True), whether to compute products of features
        - verbose: verbosity level (int; default: 0)

    Returns:
        - np.array (n_datapoints x n_additional_features) with newly computed features
        - list with n_additional_features names describing the newly computed features
    """
    # check how many new features we will compute
    d = X.shape[1]
    n = 0
    if compute_ratio:
        n += d * d - d
    if compute_product:
        n += (d * d - d) // 2
    if not n:
        logging.error("[AutoFeatLight] call _compute_additional_features with compute_ratio and/or compute_product set to True")
        return None, []
    if not feature_names:
        feature_names = [f"x{i}" for i in range(1, d + 1)]
    # compute new features
    if verbose:
        logging.info(f"[AutoFeatLight] computing {n} additional features from {d} original features")
    new_features = []
    X_new = np.zeros((X.shape[0], n))
    new_i = 0
    if compute_ratio:
        for i in range(d):
            # compute 1/x1
            with np.errstate(divide="ignore"):
                x = 1 / X[:, i]
            # instead of dividing by 0 for some data points we just set the new feature to 0 there
            x[np.invert(np.isfinite(x))] = 0.0
            for j in range(d):
                if i != j:
                    # multiply with x2 to get x2/x1
                    X_new[:, new_i] = x * X[:, j]
                    new_features.append(f"{feature_names[j]} / {feature_names[i]}")
                    new_i += 1
    if compute_product:
        for i in range(d):
            for j in range(i + 1, d):
                X_new[:, new_i] = X[:, i] * X[:, j]
                new_features.append(f"{feature_names[i]} * {feature_names[j]}")
                new_i += 1
    assert new_i == n, f"Internal Error in _compute_additional_features: new_i: {new_i}, n: {n}"
    return X_new, new_features


class AutoFeatLight(BaseEstimator):
    def __init__(
        self,
        compute_ratio: bool = True,
        compute_product: bool = True,
        scale: bool = False,
        power_transform: bool = False,
        corrthr: float = 0.995,
        corrthr_init: float = 0.99999,
        verbose: int = 0,
    ):
        """
        Basic Feature Engineering:
            - remove useless features (zero variance or redundant)
            - compute additional non-linear features (ratios and product of original features, i.e. x1/x2 and x1*x2)
            - make all features more normally distributed (using the yeo-johnson power transform)

        Inputs:
            - compute_ratio: bool (default: True), whether to compute ratios of features
            - compute_product: bool (default: True), whether to compute products of features
            - scale: bool (default: False), rudimentary scaling of the data (only relevant if not computing the power_transform anyways)
            - power_transform: bool (default: False), whether to use a power transform (yeo-johnson) to make all features more normally distributed
            - corrthr: threshold for correlations: if a feature has a higher pearson correlation to another feature it's
                       considered as redundant and ignored (float; default: 0.995)
            - corrthr_init: correlation threshold for initial features (before the feat eng step) (float; default: 0.99999)
            - verbose: verbosity level (int; default: 0)

        Attributes (after calling fit/fit_transform):
            - features_: feature names of transformed features
            - original_columns_: original columns of X when calling fit
            - return_df_: whether fit was called with a dataframe in which case a df will be returned as well,
                          otherwise a numpy array
            - good_cols_org_: list of good features from the original inputs
            - scaler_: if scale: fitted standard scaler
            - power_transformer_: if power_transform: fitted power transform
        """
        self.compute_ratio = compute_ratio
        self.compute_product = compute_product
        self.scale = scale
        self.power_transform = power_transform
        self.corrthr_init = corrthr_init
        self.corrthr = corrthr
        self.verbose = verbose

    def fit(self, X: np.ndarray | pd.DataFrame):
        """
        WARNING: call fit_transform instead!

        Inputs:
            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)
        """
        if self.verbose:
            logging.warning("[AutoFeatLight] This just calls fit_transform() but does not return the transformed dataframe.")
            logging.info("[AutoFeatLight] It is much more efficient to call fit_transform() instead of fit() and transform()!")
        _ = self.fit_transform(X)
        return self

    def transform(self, X: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
        """
        Inputs:
            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)

        Returns:
            - new_X: new pandas dataframe or numpy array with additional/transformed features
        """
        check_is_fitted(self, ["good_cols_org_"])
        if not self.good_cols_org_:
            if self.verbose > 0:
                logging.warning("[AutoFeatLight] No good features found; returning data unchanged.")
            return X
        if isinstance(X, pd.DataFrame):
            # make sure all data is numeric or we'll get an error when checking X
            X = X.select_dtypes(include=np.number)
            df_index = X.index
        else:
            df_index = None
        # check input
        cols = list(X.columns) if isinstance(X, pd.DataFrame) else [f"x{i}" for i in range(1, X.shape[1] + 1)]
        X = check_array(X, force_all_finite="allow-nan")
        if not cols == self.original_columns_:
            raise ValueError("[AutoFeatLight] Not the same features as when calling fit.")
        # sort out useless original columns
        df = pd.DataFrame(X, columns=cols, index=df_index)[self.good_cols_org_]
        if self.compute_ratio or self.compute_product:
            # compute additional useful features
            X_new, new_features = _compute_additional_features(
                df.to_numpy(),
                self.good_cols_org_,
                self.compute_ratio,
                self.compute_product,
                self.verbose,
            )
            df = pd.concat([df, pd.DataFrame(X_new, columns=new_features)], axis=1)
            df = df[self.features_]
        # scale/transform
        if self.scale or self.power_transform:
            X_new = self.scaler_.transform(df.to_numpy())
            if self.power_transform:
                X_new = self.power_transformer_.transform(X_new)
            df = pd.DataFrame(X_new, columns=df.columns, index=df.index)
        # return either dataframe or array
        return df if self.return_df_ else df.to_numpy()

    def fit_transform(self, X: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
        """
        Inputs:
            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)

        Returns:
            - new_X: new pandas dataframe or numpy array with additional/transformed features
        """
        self.return_df_ = isinstance(X, pd.DataFrame)
        if isinstance(X, pd.DataFrame):
            # make sure all data is numeric or we'll get an error when checking X
            X = X.select_dtypes(include=np.number)
            df_index = X.index
        else:
            df_index = None
        # store column names as they'll be lost in the other check
        self.original_columns_ = list(X.columns) if isinstance(X, pd.DataFrame) else [f"x{i}" for i in range(1, X.shape[1] + 1)]
        # check input
        X = check_array(X, force_all_finite="allow-nan")
        # transform X into a dataframe (again)
        df = pd.DataFrame(X, columns=self.original_columns_, index=df_index)
        # see which of the original features are not completely useless
        self.good_cols_org_ = _check_features(df, self.corrthr_init, self.verbose)
        if not self.good_cols_org_:
            if self.verbose > 0:
                logging.warning("[AutoFeatLight] No good features found; returning original features.")
            return df if self.return_df_ else X
        # compute additional features
        df = df[self.good_cols_org_]
        if self.compute_ratio or self.compute_product:
            X_new, new_features = _compute_additional_features(
                df.to_numpy(),
                self.good_cols_org_,
                self.compute_ratio,
                self.compute_product,
                self.verbose,
            )
            # add new features to original dataframe
            df = pd.concat([df, pd.DataFrame(X_new, columns=new_features, index=df_index)], axis=1)
            # check again which of the features we should keep
            self.features_ = _check_features(df, self.corrthr, self.verbose)
            df = df[self.features_]
        else:
            self.features_ = self.good_cols_org_
        if self.scale or self.power_transform:
            # scale data to avoid errors due to large numbers
            self.scaler_ = StandardScaler(with_mean=False)
            X_new = self.scaler_.fit_transform(df.to_numpy())
            if self.power_transform:
                self.power_transformer_ = PowerTransformer(method="yeo-johnson", standardize=True)
                X_new = self.power_transformer_.fit_transform(X_new)
            df = pd.DataFrame(X_new, columns=df.columns, index=df.index)
        if self.verbose > 0:
            logging.info(f"[AutoFeatLight] New data shape: {df.shape[0]} x {df.shape[1]}")
        self.n_features_in_ = len(self.original_columns_)
        # return either dataframe or array
        return df if self.return_df_ else df.to_numpy()



# Author: Jeethu Rao <jboloor@acm.org>
# License: MIT

from __future__ import annotations

import numba as nb
import numpy as np


@nb.njit(inline="always")
def nb_apply_along_axis(func1d, axis, arr):
    assert arr.ndim == 2
    assert axis in [0, 1]
    if axis == 0:
        result = np.empty(arr.shape[1])
        for i in range(len(result)):
            result[i] = func1d(arr[:, i])
    else:
        result = np.empty(arr.shape[0])
        for i in range(len(result)):
            result[i] = func1d(arr[i, :])
    return result


@nb.njit(cache=True)
def nb_nanmean(array, axis):
    return nb_apply_along_axis(np.nanmean, axis, array)


@nb.njit(cache=True)
def nb_nanstd(array, axis):
    return nb_apply_along_axis(np.nanstd, axis, array)


@nb.njit(cache=True)
def nb_standard_scale(array):
    return (array - nb_nanmean(array, 0)) / nb_nanstd(array, 0)



# Author: Franziska Horn <cod3licious@gmail.com>
# License: MIT

from __future__ import annotations

import logging
import operator as op
import re
from functools import reduce
from itertools import combinations, product
from typing import Callable

import numba as nb
import numpy as np
import pandas as pd
import pint
import sympy
from sklearn.preprocessing import StandardScaler
from sympy.utilities.lambdify import lambdify

logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s", level=logging.INFO)


def colnames2symbols(c: str | int, i: int = 0) -> str:
    # take a messy column name and transform it to something sympy can handle
    # worst case: i is the number of the features
    # has to be a string
    c = str(c)
    # should not contain non-alphanumeric characters
    c = re.sub(r"\W+", "", c)
    if not c:
        c = f"x{i:03}"
    elif c[0].isdigit():
        c = "x" + c
    return c


def ncr(n: int, r: int) -> int:
    # compute number of combinations for n chose r
    r = min(r, n - r)
    numer = reduce(op.mul, range(n, n - r, -1), 1)
    denom = reduce(op.mul, range(1, r + 1), 1)
    return numer // denom


def n_cols_generated(n_features: int, max_steps: int, n_transformations: int = 7, n_combinations: int = 4) -> int:
    """
    computes the upper bound of how many features will be generated based on n_features to start with
    and max_steps feateng steps.
    """
    # n_transformations is 1-len(func_transform) because either abs() or sqrt and log will be applied
    n_transformations -= 1
    original_cols = n_features
    new_cols = 0
    new_new_cols = 0
    # count additions at the highest level
    n_additions = 0
    steps = 1
    if steps <= max_steps:
        # Step 1: apply transformations to original features
        original_cols += n_features * n_transformations
        # n_additions += n_features * 2  # only if 1+ or 1- is in transformations!
        steps += 1
    if steps <= max_steps:
        # Step 2: first combination of features
        new_cols = n_combinations * (ncr(original_cols, 2))
        n_additions += 3 * new_cols // 4
        steps += 1
    while steps <= max_steps:
        # apply transformations on these new features
        # n_additions += new_cols * 2
        new_cols += new_cols * n_transformations
        steps += 1
        # get combinations of old and new features
        if steps <= max_steps:
            new_new_cols = n_combinations * (original_cols * new_cols)
            n_additions += 3 * new_new_cols // 4
            steps += 1
        # and combinations of new features within themselves
        if steps <= max_steps:
            n = n_combinations * (ncr(new_cols, 2))
            new_new_cols += n
            n_additions += 3 * n // 4
            steps += 1
            # update old and new features and repeat
            original_cols += new_cols
            new_cols = new_new_cols
            new_new_cols = 0
    # finally, apply transformation on the last new features
    if steps <= max_steps:
        # n_additions += new_cols * 2
        new_cols += new_cols * n_transformations
    return original_cols + new_cols + new_new_cols - n_additions


def engineer_features(
    df_org: pd.DataFrame,
    start_features: list | None = None,
    units: dict | None = None,
    max_steps: int = 3,
    transformations: list | tuple = ("1/", "exp", "log", "abs", "sqrt", "^2", "^3"),
    verbose: int = 0,
) -> tuple[pd.DataFrame, dict]:
    """
    Given a DataFrame with original features, perform the feature engineering routine for max_steps.
    It starts with a transformation of the original features (applying log, ^2, sqrt, etc.),
    then in the next step, the features are combined (x+y, x*y, ...), and in further steps, the resulting
    features are again transformed and combinations of the resulting features are computed.

    Inputs:
        - df_org: pandas DataFrame with original features in columns
        - start_features: list with column names for df_org with features that should be considered for expansion
                          (default: None --> all columns)
        - units: a dict with {column_name: pint.Quantity}: some operations like x+y can only be performed if the
                 features have comparable units (default None: all combinations are allowed)
                 careful: will be modified in place!
        - max_steps: how many feature engineering steps should be performed. Default is 3, this produces:
            Step 1: transformation of original features
            Step 2: first combination of features
            Step 3: transformation of new features
            (Step 4: combination of old and new features)
            --> with 3 original features, after 4 steps you will already end up with around 200k features!
        - transformations: list of transformations that should be applied; possible elements:
                           "1/", "exp", "log", "abs", "sqrt", "^2", "^3", "1+", "1-", "sin", "cos", "exp-", "2^"
                           (first 7, i.e., up to ^3, are applied by default)
        - verbose: verbosity level (int; default: 0)

    Returns:
        - df: new DataFrame with all features in columns
        - feature_pool: dict with {col: sympy formula} formulas to generate each feature
    """
    # initialize the feature pool with columns from the dataframe
    if start_features is None:
        start_features = list(df_org.columns)
    else:
        for c in start_features:
            if c not in df_org.columns:
                raise ValueError(f"[feateng] start feature {c} not in df_org.columns")
    feature_pool = {c: sympy.symbols(colnames2symbols(c, i), real=True) for i, c in enumerate(start_features)}
    if max_steps < 1:
        if verbose > 0:
            logging.warning("[feateng] no features generated for max_steps < 1.")
        return df_org, feature_pool
    # get a copy of the dataframe - this is where all the features will be added
    df = pd.DataFrame(df_org.copy(), dtype=np.float32)

    compiled_func_transformations = None
    compiled_func_transforms_cond = None
    compiled_func_combinations = None

    def compile_func_transform(name: str, ft: Callable, plus_1: bool = False):
        def _abs(x):
            return np.abs(x)

        # create temporary variable expression and apply it to precomputed feature
        t = sympy.symbols("t")
        expr_temp = ft(t + 1) if plus_1 else ft(t)
        fn = _abs if name == "abs" else lambdify(t, expr_temp)
        return nb.njit(fn)

    def apply_transformations(features_list: list) -> tuple[list, set]:
        # feature transformations
        func_transform = {
            "exp": lambda x: sympy.exp(x),
            "exp-": lambda x: sympy.exp(-x),
            "log": lambda x: sympy.log(x),
            "abs": lambda x: sympy.Abs(x),
            "sqrt": lambda x: sympy.sqrt(x),
            "sin": lambda x: sympy.sin(x),
            "cos": lambda x: sympy.cos(x),
            "2^": lambda x: 2**x,
            "^2": lambda x: x**2,
            "^3": lambda x: x**3,
            "1+": lambda x: 1 + x,
            "1-": lambda x: 1 - x,
            "1/": lambda x: 1 / x,
        }
        func_transform_units = {
            "exp": lambda x: np.exp(x),
            "exp-": lambda x: np.exp(-x),
            "log": lambda x: np.log(x),
            "abs": lambda x: np.abs(x),
            "sqrt": lambda x: np.sqrt(x),
            "sin": lambda x: np.sin(x),
            "cos": lambda x: np.cos(x),
            "2^": lambda x: np.exp(x),
            "^2": lambda x: x**2,
            "^3": lambda x: x**3,
            "1+": lambda x: 1 + x,
            "1-": lambda x: 1 - x,
            "1/": lambda x: 1 / x,
        }
        # conditions on the original features that have to be met to apply the transformation
        func_transform_cond = {
            "exp": lambda x: np.all(x < 10),
            "exp-": lambda x: np.all(-x < 10),
            "log": lambda x: np.all(x >= 0),
            "abs": lambda x: np.any(x < 0),
            "sqrt": lambda x: np.all(x >= 0),
            "sin": lambda x: True,
            "cos": lambda x: True,
            "2^": lambda x: np.all(x < 50),
            "^2": lambda x: np.all(np.abs(x) < 1000000),
            "^3": lambda x: np.all(np.abs(x) < 10000),
            "1+": lambda x: True,
            "1-": lambda x: True,
            "1/": lambda x: np.all(x != 0),
        }
        # apply transformations to the features in the given features list
        # modifies global variables df and feature_pool!
        nonlocal df, feature_pool, units
        nonlocal compiled_func_transformations, compiled_func_transforms_cond

        if compiled_func_transformations is None:
            compiled_func_transformations = {k: compile_func_transform(k, v) for k, v in func_transform.items()}
            compiled_func_transformations["log_plus_1"] = compile_func_transform("log", func_transform["log"], plus_1=True)

            compiled_func_transforms_cond = {x[0]: nb.njit(x[1]) for x in func_transform_cond.items()}

        # returns a list of new features that were generated
        new_features: list[str] = []
        uncorr_features = set()
        # store all new features in a preallocated numpy array before adding it to the dataframe
        feat_array = np.zeros((df.shape[0], len(features_list) * len(transformations)), dtype=np.float32)
        cat_features = {feat for feat in features_list if len(df[feat].unique()) <= 2}
        func_transform_cond_cache = {}  # Cache for compiled_func_transforms_cond checks
        for i, feat in enumerate(features_list):
            if verbose and not i % 100:
                print(f"[feateng] {i:15}/{len(features_list):15} features transformed", end="\r")
            for ft in transformations:
                # (don't compute transformations on categorical features)
                if feat in cat_features:
                    continue
                # check if transformation is valid for particular feature (i.e. given actual numerical values)
                cache_key = (ft, feat)
                if cache_key not in func_transform_cond_cache:
                    func_transform_cond_cache[cache_key] = compiled_func_transforms_cond[ft](df[feat].to_numpy())
                if func_transform_cond_cache[cache_key]:
                    # get the expression (based on the primary features)
                    expr = func_transform[ft](feature_pool[feat])
                    expr_name = str(expr)
                    # we're simplifying expressions, so we might already have that one
                    if expr_name not in feature_pool:
                        # if we're given units, check if the operation is legal
                        if units:
                            try:
                                units[expr_name] = func_transform_units[ft](units[feat])
                                units[expr_name].__dict__["_magnitude"] = 1.0
                            except (pint.DimensionalityError, pint.OffsetUnitCalculusError):
                                continue
                        feature_pool[expr_name] = expr
                        if expr == "log" and np.any(df[feat] < 1):
                            f = compiled_func_transformations["log_plus_1"]
                        else:
                            f = compiled_func_transformations[ft]
                        new_feat = np.array(f(df[feat].to_numpy()), dtype=np.float32)
                        # near 0 variance test - sometimes all that's left is "e"
                        if np.isfinite(new_feat).all() and np.var(new_feat) > 1e-10:
                            corr = abs(np.corrcoef(new_feat, df[feat])[0, 1])
                            if corr < 1.0:
                                feat_array[:, len(new_features)] = new_feat
                                new_features.append(expr_name)
                                # correlation test: don't include features that are basically the same as the original features
                                # but we only filter them out at the end, since they still might help in other steps!
                                if corr < 0.95:
                                    uncorr_features.add(expr_name)
        if verbose > 0:
            logging.info(
                f"[feateng] Generated {len(new_features)} transformed features from {len(features_list)} original features - done.",
            )
        df = df.join(pd.DataFrame(feat_array[:, : len(new_features)], columns=new_features, index=df.index, dtype=np.float32))
        return new_features, uncorr_features

    def compile_func_combinations(func_combinations: dict) -> dict:
        d = {}
        for fc in func_combinations:
            s, t = sympy.symbols("s t")
            expr_temp = func_combinations[fc](s, t)
            fn = lambdify((s, t), expr_temp)
            vect = nb.vectorize(["float32(float32, float32)"], nopython=True)
            d[fc] = vect(fn)
        return d

    def get_feature_combinations(feature_tuples: list) -> tuple[list, set]:
        # new features as combinations of two other features
        func_combinations = {
            "x+y": lambda x, y: x + y,
            "x*y": lambda x, y: x * y,
            "x-y": lambda x, y: x - y,
            "y-x": lambda x, y: y - x,
        }
        # get all feature combinations for the given feature tuples
        # modifies global variables df and feature_pool!
        nonlocal df, feature_pool, units, compiled_func_combinations

        if compiled_func_combinations is None:
            compiled_func_combinations = compile_func_combinations(func_combinations)

        # only compute all combinations if there are more transformations applied afterwards
        # additions at the highest level are sorted out later anyways
        combinations = ["x*y"] if steps == max_steps else list(func_combinations.keys())
        # returns a list of new features that were generated
        new_features: list[str] = []
        uncorr_features = set()
        # store all new features in a preallocated numpy array before adding it to the dataframe
        feat_array = np.zeros((df.shape[0], len(feature_tuples) * len(combinations)), dtype=np.float32)
        for i, (feat1, feat2) in enumerate(feature_tuples):
            if verbose and not i % 100:
                print(f"[feateng] {i:15}/{len(feature_tuples):15} feature tuples combined", end="\r")
            for fc in combinations:
                expr = func_combinations[fc](feature_pool[feat1], feature_pool[feat2])
                expr_name = str(expr)
                if expr_name not in feature_pool:
                    # if we're given units, check if the operation is legal
                    if units:
                        try:
                            units[expr_name] = func_combinations[fc](units[feat1], units[feat2])
                            units[expr_name].__dict__["_magnitude"] = 1.0
                        except (pint.DimensionalityError, pint.OffsetUnitCalculusError):
                            continue
                    feature_pool[expr_name] = expr
                    f = compiled_func_combinations[fc]
                    new_feat = np.array(f(df[feat1].to_numpy(), df[feat2].to_numpy()), dtype=np.float32)
                    # near 0 variance test - sometimes all that's left is "e"
                    if np.isfinite(new_feat).all() and np.var(new_feat) > 1e-10:
                        corr = max(abs(np.corrcoef(new_feat, df[feat1])[0, 1]), abs(np.corrcoef(new_feat, df[feat2])[0, 1]))
                        if corr < 1.0:
                            feat_array[:, len(new_features)] = new_feat
                            new_features.append(expr_name)
                            # correlation test: don't include features that are basically the same as the original features
                            # but we only filter them out at the end, since they still might help in other steps!
                            if corr < 0.95:
                                uncorr_features.add(expr_name)
        if verbose > 0:
            logging.info(
                f"[feateng] Generated {len(new_features)} feature combinations from {len(feature_tuples)} original feature tuples - done.",
            )
        df = df.join(pd.DataFrame(feat_array[:, : len(new_features)], columns=new_features, index=df.index, dtype=np.float32))
        return new_features, uncorr_features

    # get transformations of initial features
    steps = 1
    if verbose > 0:
        logging.info("[feateng] Step 1: transformation of original features")
    original_features = list(feature_pool.keys())
    uncorr_features = set(feature_pool.keys())
    temp_new, temp_uncorr = apply_transformations(original_features)
    original_features.extend(temp_new)
    uncorr_features.update(temp_uncorr)
    steps += 1
    # get combinations of first feature set
    if steps <= max_steps:
        if verbose > 0:
            logging.info("[feateng] Step 2: first combination of features")
        new_features, temp_uncorr = get_feature_combinations(list(combinations(original_features, 2)))
        uncorr_features.update(temp_uncorr)
        steps += 1
    while steps <= max_steps:
        # apply transformations on these new features
        if verbose > 0:
            logging.info(f"[feateng] Step {steps}: transformation of new features")
        temp_new, temp_uncorr = apply_transformations(new_features)
        new_features.extend(temp_new)
        uncorr_features.update(temp_uncorr)
        steps += 1
        # get combinations of old and new features
        if steps <= max_steps:
            if verbose > 0:
                logging.info(f"[feateng] Step {steps}: combining old and new features")
            new_new_features, temp_uncorr = get_feature_combinations(list(product(original_features, new_features)))
            uncorr_features.update(temp_uncorr)
            steps += 1
        # and combinations of new features within themselves
        if steps <= max_steps:
            if verbose > 0:
                logging.info(f"[feateng] Step {steps}: combining new features")
            temp_new, temp_uncorr = get_feature_combinations(list(combinations(new_features, 2)))
            new_new_features.extend(temp_new)
            uncorr_features.update(temp_uncorr)
            steps += 1
            # update old and new features and repeat
            original_features.extend(new_features)
            new_features = new_new_features

    # sort out all features that are just additions on the highest level or correlated with more basic features
    if verbose > 0:
        logging.info(
            f"[feateng] Generated altogether {len(feature_pool) - len(start_features)} new features in {max_steps} steps"
        )
        logging.info("[feateng] Removing correlated features, as well as additions at the highest level")
    feature_pool = {
        c: feature_pool[c] for c in feature_pool if c in uncorr_features and feature_pool[c].func != sympy.core.add.Add
    }
    cols = [
        c for c in list(df.columns) if c in feature_pool and c not in df_org.columns
    ]  # categorical cols not in feature_pool
    if cols:
        # check for correlated features again; this time with the start features
        corrs = dict(
            zip(
                cols,
                np.max(
                    np.abs(
                        np.dot(StandardScaler().fit_transform(df[cols]).T, StandardScaler().fit_transform(df_org))
                        / df_org.shape[0],
                    ),
                    axis=1,
                ),
            ),
        )
        cols = [c for c in cols if corrs[c] < 0.9]
    cols = list(df_org.columns) + cols
    if verbose > 0:
        logging.info(f"[feateng] Generated a total of {len(feature_pool) - len(start_features)} additional features")
    return df[cols], feature_pool



# Author: Franziska Horn <cod3licious@gmail.com>
# License: MIT

from __future__ import annotations

import logging
import warnings
from collections import Counter

import numpy as np
import pandas as pd
import sklearn.linear_model as lm
from joblib import Parallel, delayed
from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y

from autofeat.nb_utils import nb_standard_scale

logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s", level=logging.INFO)


def _add_noise_features(X: np.ndarray):
    """
    Adds 3-1.5*d additional noise features to X.

    Inputs:
        - X: n x d numpy array with d features
    Returns:
        - X with additional noise features
    """
    n_feat = X.shape[1]
    if X.shape[0] > 50 and n_feat > 1:
        # shuffled features
        rand_noise = nb_standard_scale(np.random.permutation(X.flatten()).reshape(X.shape))
        X = np.hstack([X, rand_noise])
    # normally distributed noise
    rand_noise = np.random.randn(X.shape[0], max(3, int(0.5 * n_feat)))
    return np.hstack([X, rand_noise])


def _noise_filtering(
    X: np.ndarray,
    target: np.ndarray,
    good_cols: list | None = None,
    problem_type: str = "regression",
) -> list:
    """
    Trains a prediction model with additional noise features and selects only those of the
    original features that have a higher coefficient than any of the noise features.

    Inputs:
        - X: n x d numpy array with d features
        - target: n dimensional array with targets corresponding to the data points in X
        - good_cols: list of column names for the features in X
        - problem_type: str, either "regression" or "classification" (default: "regression")

    Returns:
        - good_cols: list of noise filtered column names
    """
    n_feat = X.shape[1]
    if good_cols is None or not len(good_cols):
        good_cols = list(range(n_feat))
    assert len(good_cols) == n_feat, "fewer column names provided than features in X."
    # perform noise filtering on these features
    if problem_type == "regression":
        model = lm.LassoLarsCV(cv=5, eps=1e-8)
    elif problem_type == "classification":
        model = lm.LogisticRegressionCV(cv=5, penalty="l1", solver="saga", class_weight="balanced")
    else:
        logging.warning(f"[featsel] Unknown problem_type {problem_type} - not performing noise filtering.")
        model = None
    if model is not None:
        X = _add_noise_features(X)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # TODO: remove if sklearn least_angle issue is fixed
            try:
                model.fit(X, target)
            except ValueError:
                rand_idx = np.random.permutation(X.shape[0])
                model.fit(X[rand_idx], target[rand_idx])
            # model.fit(X, target)
        # for classification, model.coefs_ is n_classes x n_features, but we need n_features
        coefs = np.abs(model.coef_) if problem_type == "regression" else np.max(np.abs(model.coef_), axis=0)
        weights = dict(zip(good_cols, coefs[: len(good_cols)]))
        # only include features that are more important than our known noise features
        noise_w_thr = np.max(coefs[n_feat:])
        good_cols = [c for c in good_cols if weights[c] > noise_w_thr]
    return good_cols


def _select_features_1run(df: pd.DataFrame, target: np.ndarray, problem_type: str = "regression", verbose: int = 0) -> list:
    """
    One feature selection run.

    Inputs:
        - df: nxp pandas DataFrame with n data points and p features; to avoid overfitting, only provide data belonging
              to the n training data points. The variables have to be scaled to have 0 mean and unit variance.
        - target: n dimensional array with targets corresponding to the data points in df
        - problem_type: str, either "regression" or "classification" (default: "regression")
        - verbose: verbosity level (int; default: 0)

    Returns:
        - good_cols: list of column names for df with which a prediction model can be trained
    """
    if df.shape[0] <= 1:
        raise ValueError(f"n_samples = {df.shape[0]}")
    # initial selection of too few but (hopefully) relevant features
    if problem_type == "regression":
        model = lm.LassoLarsCV(cv=5, eps=1e-8)
    elif problem_type == "classification":
        model = lm.LogisticRegressionCV(cv=5, penalty="l1", solver="saga", class_weight="balanced")
    else:
        logging.warning(f"[featsel] Unknown problem_type {problem_type} - not performing feature selection!")
        return []
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # TODO: remove if sklearn least_angle issue is fixed
        try:
            model.fit(df, target)
        except ValueError:
            # try once more with shuffled data, if it still doesn't work, give up
            rand_idx = np.random.permutation(df.shape[0])
            model.fit(df.iloc[rand_idx], target[rand_idx])
        # model.fit(df, target)
    # for classification, model.coefs_ is n_classes x n_features, but we need n_features
    coefs = np.abs(model.coef_) if problem_type == "regression" else np.max(np.abs(model.coef_), axis=0)
    # weight threshold: select at most 0.2*n_train initial features
    thr = sorted(coefs, reverse=True)[min(df.shape[1] - 1, df.shape[0] // 5)]
    initial_cols = list(df.columns[coefs > thr])
    # noise filter
    initial_cols = _noise_filtering(df[initial_cols].to_numpy(), target, initial_cols, problem_type)
    good_cols_set = set(initial_cols)
    if verbose > 0:
        logging.info(f"[featsel]\t {len(initial_cols)} initial features.")
    # add noise features
    X_w_noise = _add_noise_features(df[initial_cols].to_numpy())
    # go through all remaining features in splits of n_feat <= 0.5*n_train
    other_cols = list(np.random.permutation(list(set(df.columns).difference(initial_cols))))
    if other_cols:
        n_splits = int(np.ceil(len(other_cols) / max(10, 0.5 * df.shape[0] - len(initial_cols))))
        split_size = int(np.ceil(len(other_cols) / n_splits))
        for i in range(n_splits):
            current_cols = other_cols[i * split_size : min(len(other_cols), (i + 1) * split_size)]
            X = np.hstack([df[current_cols].to_numpy(), X_w_noise])
            if problem_type == "regression":
                model = lm.LassoLarsCV(cv=5, eps=1e-8)
            else:
                model = lm.LogisticRegressionCV(cv=5, penalty="l1", solver="saga", class_weight="balanced")
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                # TODO: remove if sklearn least_angle issue is fixed
                try:
                    model.fit(X, target)
                except ValueError:
                    rand_idx = np.random.permutation(X.shape[0])
                    model.fit(X[rand_idx], target[rand_idx])
                # model.fit(X, target)
            current_cols.extend(initial_cols)
            # for classification, model.coefs_ is n_classes x n_features, but we need n_features
            coefs = np.abs(model.coef_) if problem_type == "regression" else np.max(np.abs(model.coef_), axis=0)
            weights = dict(zip(current_cols, coefs[: len(current_cols)]))
            # only include features that are more important than our known noise features
            noise_w_thr = np.max(coefs[len(current_cols) :])
            good_cols_set.update([c for c in weights if abs(weights[c]) > noise_w_thr])
            if verbose > 0:
                print(
                    f"[featsel]\t Split {i + 1:2}/{n_splits}: {len(good_cols_set):3} candidate features identified.",
                    end="\r",
                )
    # noise filtering on the combination of features
    good_cols = list(good_cols_set)
    good_cols = _noise_filtering(df[good_cols].to_numpy(), target, good_cols, problem_type)
    if verbose > 0:
        logging.info(f"\n[featsel]\t Selected {len(good_cols):3} features after noise filtering.")
    return good_cols


def select_features(
    df: pd.DataFrame,
    target: np.ndarray,
    featsel_runs: int = 5,
    keep: list | None = None,
    problem_type: str = "regression",
    n_jobs: int = 1,
    verbose: int = 0,
) -> list:
    """
    Selects predictive features given the data and targets.

    Inputs:
        - df: nxp pandas DataFrame with n data points and p features; to avoid overfitting, only provide data belonging
              to the n training data points.
        - target: n dimensional array with targets corresponding to the data points in df
        - featsel_runs: number of times to perform in the feature selection part with a random fraction of data points (int; default: 5)
        - keep: list of features that should be kept no matter what
        - problem_type: str, either "regression" or "classification" (default: "regression")
        - n_jobs: how many jobs to run when selecting the features in parallel (int; default: 1)
        - verbose: verbosity level (int; default: 0)

    Returns:
        - good_cols: list of column names for df with which a regression model can be trained
    """
    if not (len(df) == len(target)):
        raise ValueError("[featsel] df and target dimension mismatch.")
    if keep is None:
        keep = []
    # check that keep columns are actually in df (- the columns might have been transformed to strings!)
    keep = [c for c in keep if c in df.columns and str(c) not in df.columns] + [str(c) for c in keep if str(c) in df.columns]
    # scale features to have 0 mean and unit std
    if verbose > 0:
        if featsel_runs > df.shape[0]:
            logging.warning("[featsel] Less data points than featsel runs!!")
        print("[featsel] Scaling data...", end="")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        df_scaled = pd.DataFrame(nb_standard_scale(df.to_numpy()), columns=df.columns, dtype=np.float32)
        target_scaled = nb_standard_scale(target.reshape(-1, 1)).ravel() if problem_type == "regression" else target
    if verbose > 0:
        print("done.")

    good_cols = list(df.columns)

    # select good features in k runs in parallel
    # by doing sort of a cross-validation (i.e., randomly subsample data points)
    def run_select_features(i: int):
        if verbose > 0:
            logging.info(f"[featsel] Feature selection run {i + 1}/{featsel_runs}")
        np.random.seed(i)
        rand_idx = np.random.permutation(df_scaled.index)[: max(10, int(0.85 * len(df_scaled)))]
        return _select_features_1run(df_scaled.iloc[rand_idx], target_scaled[rand_idx], problem_type, verbose=verbose - 1)

    if featsel_runs >= 1 and problem_type in ("regression", "classification"):
        if n_jobs == 1 or featsel_runs == 1:
            # only use parallelization code if you actually parallelize
            selected_columns = []
            for i in range(featsel_runs):
                selected_columns.extend(run_select_features(i))
        else:

            def flatten_lists(l: list):
                return [item for sublist in l for item in sublist]

            selected_columns = flatten_lists(
                Parallel(n_jobs=n_jobs, verbose=100 * verbose)(delayed(run_select_features)(i) for i in range(featsel_runs)),
            )

        if selected_columns:
            selected_columns_counter = Counter(selected_columns)
            # sort by frequency, but down weight longer formulas to break ties
            selected_columns = sorted(
                selected_columns_counter,
                key=lambda x: selected_columns_counter[x] - 0.000001 * len(str(x)),
                reverse=True,
            )
            if verbose > 0:
                logging.info(f"[featsel] {len(selected_columns)} features after {featsel_runs} feature selection runs")
            # correlation filtering
            selected_columns = keep + [c for c in selected_columns if c not in keep]
            if not keep:
                good_cols = [selected_columns[0]]
                k = 1
            else:
                good_cols = keep
                k = len(keep)
            if len(selected_columns) > k:
                correlations = df_scaled[selected_columns].corr()
                for i, c in enumerate(selected_columns[k:], k):
                    # only take features that are somewhat uncorrelated with the rest
                    if np.max(np.abs(correlations[c].ravel()[:i])) < 0.9:
                        good_cols.append(c)
            if verbose > 0:
                logging.info(f"[featsel] {len(good_cols)} features after correlation filtering")

    # perform noise filtering on these features
    good_cols = _noise_filtering(df_scaled[good_cols].to_numpy(), target_scaled, good_cols, problem_type)
    if verbose > 0:
        logging.info(f"[featsel] {len(good_cols)} features after noise filtering")
        if not good_cols:
            logging.warning("[featsel] Not a single good features was found...")

    # add keep columns back in
    good_cols = keep + [c for c in good_cols if c not in keep]
    if verbose > 0 and keep:
        logging.info(f"[featsel] {len(good_cols)} final features selected (including {len(keep)} original keep features).")
    return good_cols


class FeatureSelector(BaseEstimator):
    def __init__(
        self,
        problem_type: str = "regression",
        featsel_runs: int = 5,
        keep: list | None = None,
        n_jobs: int = 1,
        verbose: int = 0,
    ):
        """
        multi-step cross-validated feature selection

        Inputs:
            - problem_type: str, either "regression" or "classification" (default: "regression")
            - featsel_runs: number of times to perform in the feature selection part with a random fraction of data points (int; default: 5)
            - keep: list of features that should be kept no matter what
            - n_jobs: how many jobs to run when selecting the features in parallel (int; default: 1)
            - verbose: verbosity level (int; default: 0)

        Attributes:
            - good_cols_: list of good features (to select via pandas DataFrame columns)
            - original_columns_: original columns of X when calling fit
            - return_df_: whether fit was called with a dataframe in which case a df will be returned as well,
                          otherwise a numpy array
        """
        self.problem_type = problem_type
        self.featsel_runs = featsel_runs
        self.keep = keep
        self.n_jobs = n_jobs
        self.verbose = verbose

    def fit(self, X: np.ndarray | pd.DataFrame, y: np.ndarray | pd.DataFrame):
        """
        Selects features.

        Inputs:
            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)
            - y: pandas dataframe or numpy array with targets for all n_datapoints
        """
        self.return_df_ = isinstance(X, pd.DataFrame)
        # store column names as they'll be lost in the other check
        # first calling np.array assures that all the column names have the same dtype
        # as otherwise we get problems when calling np.random.permutation on the columns
        cols = list(np.array(list(X.columns))) if isinstance(X, pd.DataFrame) else []
        # check input variables
        X, target = check_X_y(X, y, y_numeric=self.problem_type == "regression")
        if not cols:
            cols = [f"x{i}" for i in range(X.shape[1])]
        self.original_columns_ = cols
        # transform X into a dataframe (again)
        df = pd.DataFrame(X, columns=cols)
        # do the feature selection
        self.good_cols_ = select_features(
            df,
            target,
            self.featsel_runs,
            self.keep,
            self.problem_type,
            self.n_jobs,
            self.verbose,
        )
        self.n_features_in_ = X.shape[1]
        return self

    def transform(self, X: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
        """
        Inputs:
            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)

        Returns:
            - new_X: new pandas dataframe or numpy array with only the good features
        """
        check_is_fitted(self, ["good_cols_"])
        if not self.good_cols_:
            if self.verbose > 0:
                logging.warning("[FeatureSelector] No good features found; returning data unchanged.")
            return X
        # store column names as they'll be lost in the other check
        # first calling np.array assures that all the column names have the same dtype
        # as otherwise we get problems when calling np.random.permutation on the columns
        cols = list(np.array(list(X.columns))) if isinstance(X, pd.DataFrame) else []
        # check input variables
        X = check_array(X, force_all_finite="allow-nan")
        if not cols:
            cols = [f"x{i}" for i in range(X.shape[1])]
        if not cols == self.original_columns_:
            raise ValueError("[FeatureSelector] Not the same features as when calling fit.")
        # transform X into a dataframe (again) and select columns
        new_X = pd.DataFrame(X, columns=cols)[self.good_cols_]
        # possibly transform into a numpy array
        if not self.return_df_:
            new_X = new_X.to_numpy()
        return new_X

    def fit_transform(self, X: np.ndarray | pd.DataFrame, y: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
        """
        Selects features and returns only those selected.

        Inputs:
            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)
            - y: pandas dataframe or numpy array with targets for all n_datapoints
        Returns:
            - new_X: new pandas dataframe or numpy array with only the good features
        """
        self.fit(X, y)
        return self.transform(X)



# Author: Franziska Horn <cod3licious@gmail.com>
# License: MIT

name = "autofeat"
__version__ = "2.1.3"
from .autofeatlight import AutoFeatLight  # noqa
from .autofeat import AutoFeatModel, AutoFeatRegressor, AutoFeatClassifier  # noqa
from .featsel import FeatureSelector  # noqa



# Author: Franziska Horn <cod3licious@gmail.com>
# License: MIT

from __future__ import annotations

import logging
import warnings

import numba as nb
import numpy as np
import pandas as pd
import pint
import sklearn.linear_model as lm
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.preprocessing import OneHotEncoder
from sklearn.utils.validation import check_array, check_is_fitted, check_X_y
from sympy.utilities.lambdify import lambdify

from autofeat.feateng import colnames2symbols, engineer_features, n_cols_generated
from autofeat.featsel import select_features

logging.basicConfig(format="%(asctime)s %(levelname)s: %(message)s", level=logging.INFO)


def _parse_units(units: dict, ureg: pint.UnitRegistry | None = None, verbose: int = 0):
    """
    Convert a dict with string units to pint quantities.

    Inputs:
        - units: dict with {"variable_name": "unit"}
        - ureg: optional: a pint UnitRegistry
        - verbose: verbosity level (int; default: 0)

    Returns:
        - parsed_units: dict with {"variable_name": pint Quantity}
    """
    parsed_units = {}
    if units:
        if ureg is None:
            ureg = pint.UnitRegistry(auto_reduce_dimensions=True, autoconvert_offset_to_baseunit=True)
        for c in units:
            try:
                parsed_units[c] = ureg.parse_expression(units[c])
            except pint.UndefinedUnitError:
                if verbose > 0:
                    logging.warning(f"[AutoFeat] unit {units[c]} of column {c} was not recognized and will be ignored!")
                parsed_units[c] = ureg.parse_expression("")
            parsed_units[c].__dict__["_magnitude"] = 1.0
    return parsed_units


class AutoFeatModel(BaseEstimator):
    def __init__(
        self,
        problem_type: str = "regression",
        categorical_cols: list | None = None,
        feateng_cols: list | None = None,
        units: dict | None = None,
        feateng_steps: int = 2,
        featsel_runs: int = 5,
        max_gb: int | None = None,
        transformations: list | tuple = ("1/", "exp", "log", "abs", "sqrt", "^2", "^3"),
        apply_pi_theorem: bool = True,
        always_return_numpy: bool = False,
        n_jobs: int = 1,
        verbose: int = 0,
    ):
        """
        multi-step feature engineering and cross-validated feature selection to generate promising additional
        features for your dataset and train a linear prediction model with them.

        Inputs:
            - problem_type: str, either "regression" or "classification" (default: "regression")
            - categorical_cols: list of column names of categorical features; these will be transformed into
                                0/1 encoding (default: None)
            - feateng_cols: list of column names that should be used for the feature engineering part
                            (default None --> all, with categorical_cols in 0/1 encoding)
            - units: dictionary with {col_name: unit} where unit is a string that can be converted into a pint unit.
                     all columns without units are dimensionless and can be combined with any other column.
                     Note: it is assumed that all features are of comparable magnitude, i.e., not one variable is in
                           m and another in mm. If this needs to be accounted for, please scale your variables before
                           passing them to autofeat!
                     (default: None --> all columns are dimensionless).
            - feateng_steps: number of steps to perform in the feature engineering part (int; default: 2)
            - featsel_runs: number of times to perform in the feature selection part with a random fraction of data points (int; default: 5)
            - max_gb: if an int is given: maximum number of gigabytes to use in the process (i.e. mostly the
                      feature engineering part). this is no guarantee! it will lead to subsampling of the
                      data points if the new dataframe generated is n_rows * n_cols * 32bit > max_gb
                      Note: this is only an approximate estimate of the final matrix; intermediate representations could easily
                            take up at least 2 or 3 times that much space...If you can, subsample before, you know your data best.
            - transformations: list of transformations that should be applied; possible elements:
                               "1/", "exp", "log", "abs", "sqrt", "^2", "^3", "1+", "1-", "sin", "cos", "exp-", "2^"
                               (first 7, i.e., up to ^3, are applied by default)
            - apply_pi_theorem: whether or not to apply the pi theorem (if units are given; bool; default: True)
            - always_return_numpy: whether to always return a numpy array instead of a pd dataframe when calling (fit_)transform
                                   (default: False; mainly used for sklearn estimator checks)
            - n_jobs: how many jobs to run when selecting the features in parallel (int; default: 1)
            - verbose: verbosity level (int; default: 0)

        Attributes:
            - original_columns_: original columns of X when calling fit
            - all_columns_: columns of X after calling fit
            - categorical_cols_map_: dict mapping from the original categorical columns to a list with new column names
            - feateng_cols_: actual columns used for the feature engineering
            - feature_formulas_: sympy formulas to generate new features
            - feature_functions_: compiled feature functions with columns
            - new_feat_cols_: list of good new features that should be generated when calling transform()
            - good_cols_: columns selected in the feature selection process, used with the final prediction model
            - prediction_model_: sklearn model instance used for the predictions

        Note: when giving categorical_cols or feateng_cols, X later (i.e. when calling fit/fit_transform) has to be a DataFrame
        """
        self.problem_type = problem_type
        self.categorical_cols = categorical_cols
        self.feateng_cols = feateng_cols
        self.units = units
        self.feateng_steps = feateng_steps
        self.max_gb = max_gb
        self.featsel_runs = featsel_runs
        self.transformations = transformations
        self.apply_pi_theorem = apply_pi_theorem
        self.always_return_numpy = always_return_numpy
        self.n_jobs = n_jobs
        self.verbose = verbose

    def __getstate__(self):
        """Get dict for pickling without feature_functions as they are not pickleable"""
        return {k: self.__dict__[k] if k != "feature_functions_" else {} for k in self.__dict__}

    def _transform_categorical_cols(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Transform categorical features into 0/1 encoding.

        Inputs:
            - df: pandas dataframe with original features
        Returns:
            - df: dataframe with categorical features transformed into multiple 0/1 columns
        """
        self.categorical_cols_map_ = {}
        if self.categorical_cols:
            e = OneHotEncoder(sparse_output=False, categories="auto")
            for c in self.categorical_cols:
                if c not in df.columns:
                    raise ValueError(f"[AutoFeat] categorical_col {c} not in df.columns")
                ohe = e.fit_transform(df[c].to_numpy()[:, None])
                new_cat_cols = [f"cat_{c}_{i}" for i in e.categories_[0]]
                self.categorical_cols_map_[c] = new_cat_cols
                df = df.join(pd.DataFrame(ohe, columns=new_cat_cols, index=df.index))
            # remove the categorical column from our columns to consider
            df = df.drop(columns=self.categorical_cols)
        return df

    def _apply_pi_theorem(self, df: pd.DataFrame) -> pd.DataFrame:
        if self.apply_pi_theorem and self.units:
            ureg = pint.UnitRegistry(auto_reduce_dimensions=True, autoconvert_offset_to_baseunit=True)
            parsed_units = _parse_units(self.units, ureg, self.verbose)
            # use only original features
            parsed_units = {c: parsed_units[c] for c in self.feateng_cols_ if not parsed_units[c].dimensionless}
            if self.verbose:
                logging.info("[AutoFeat] Applying the Pi Theorem")
            pi_theorem_results = ureg.pi_theorem(parsed_units)
            for i, r in enumerate(pi_theorem_results, 1):
                if self.verbose:
                    logging.info(f"[AutoFeat] Pi Theorem {i}: ")
                    logging.info(pint.formatter(r.items()))
                # compute the final result by multiplying and taking the power of
                cols = sorted(r)
                # only use data points where non of the affected columns are NaNs
                not_na_idx = df[cols].notna().all(axis=1)
                ptr = df[cols[0]].to_numpy()[not_na_idx] ** r[cols[0]]
                for c in cols[1:]:
                    ptr *= df[c].to_numpy()[not_na_idx] ** r[c]
                df.loc[not_na_idx, f"PT{i}_{pint.formatter(r.items()).replace(' ', '')}"] = ptr
        return df

    def _generate_features(self, df: pd.DataFrame, new_feat_cols: list) -> pd.DataFrame:
        """
        Generate additional features based on the feature formulas for all data points in the df.
        Only works after the model was fitted.

        Inputs:
            - df: pandas dataframe with original features
            - new_feat_cols: names of new features that should be generated (keys of self.feature_formulas_)

        Returns:
            - df: dataframe with the additional feature columns added
        """
        check_is_fitted(self, ["feature_formulas_"])
        if not new_feat_cols:
            return df
        if new_feat_cols[0] not in self.feature_formulas_:
            raise RuntimeError("[AutoFeat] First call fit or fit_transform to generate the features!")
        if self.verbose:
            logging.info(f"[AutoFeat] Computing {len(new_feat_cols)} new features.")
        # generate all good feature; unscaled this time
        feat_array = np.zeros((len(df), len(new_feat_cols)))
        for i, expr in enumerate(new_feat_cols):
            if self.verbose:
                print(f"[AutoFeat] {i:5}/{len(new_feat_cols):5} new features", end="\r")
            if expr not in self.feature_functions_:
                # generate a substitution expression based on all the original symbols of the original features
                # for the given generated feature in good cols
                # since sympy can handle only up to 32 original features in ufunctify, we need to check which features
                # to consider here, therefore perform some crude check to limit the number of features used
                cols = [c for i, c in enumerate(self.feateng_cols_) if colnames2symbols(c, i) in expr]
                if not cols:
                    # this can happen if no features were selected and the expr is "E" (i.e. the constant e)
                    f = None
                    f_jit = None
                else:
                    try:
                        f = lambdify([self.feature_formulas_[c] for c in cols], self.feature_formulas_[expr])
                        f_jit = nb.njit(f)
                    except Exception:
                        logging.exception(f"[AutoFeat] Error while processing expression: {expr}")
                        raise
                self.feature_functions_[expr] = (cols, f, f_jit)
            else:
                cols, f, f_jit = self.feature_functions_[expr]
            if f is not None and f_jit is not None:
                # only generate features for completely not-nan rows
                not_na_idx = df[cols].notna().all(axis=1)
                try:
                    try:
                        feat = f_jit(*(df[c].to_numpy(dtype=float)[not_na_idx] for c in cols))
                    except nb.TypingError:
                        # lambified abs fn with non trivial inputs doesn't jit compile with numba, yet
                        # fallback on the non jitted version of the function
                        feat = f(*(df[c].to_numpy(dtype=float)[not_na_idx] for c in cols))
                        # henceforth, always use the non jitted version of the function
                        self.feature_functions_[expr] = (cols, f, f)
                    feat_array[not_na_idx, i] = feat
                    feat_array[~not_na_idx, i] = np.nan
                except RuntimeWarning:
                    logging.warning(
                        f"[AutoFeat] Problem while evaluating expression: {expr} with columns {cols}",
                        " - is the data in a different range then when calling .fit()? Are maybe some values 0 that shouldn't be?",
                    )
                    raise
        if self.verbose:
            logging.info(f"[AutoFeat] {len(new_feat_cols):5}/{len(new_feat_cols):5} new features ...done.")
        return df.join(pd.DataFrame(feat_array, columns=new_feat_cols, index=df.index))

    def _X2df(self, X: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
        """
        Helper function that ensures correctness of the input data for classification tasks.
        Inputs:
            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)

        Returns:
            - df: transformed dataframe
        """
        # store column names as they'll be lost in the other check
        cols = [str(c) for c in X.columns] if isinstance(X, pd.DataFrame) else []
        # check input variables
        X = check_array(X, dtype=None)
        if not cols:
            cols = [f"x{i:03}" for i in range(X.shape[1])]
        # transform X into a dataframe (again)
        df = pd.DataFrame(X, columns=cols)
        # do we need to call transform?
        if list(df.columns) != self.all_columns_:
            temp = self.always_return_numpy
            self.always_return_numpy = False
            df = self.transform(df)
            self.always_return_numpy = temp
        return df

    def fit_transform(self, X: np.ndarray | pd.DataFrame, y: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
        """
        Fits the regression model and returns a new dataframe with the additional features.

        Inputs:
            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)
            - y: pandas dataframe or numpy array with targets for all n_datapoints
        Returns:
            - new_df: new pandas dataframe with all the original features (except categorical features transformed
                      into multiple 0/1 columns) and the most promising engineered features. This df can then be
                      used to train your final model.

        Please ensure that X only contains valid feature columns (including possible categorical variables).

        Note: we strongly encourage you to name your features X1 ...  Xn or something simple like this before passing
              a DataFrame to this model. This can help avoid potential problems with sympy later on.
              The data should only contain finite values (no NaNs etc.)
        """
        # store column names as they'll be lost in the other check
        cols = [str(c) for c in X.columns] if isinstance(X, pd.DataFrame) else []
        # check input variables
        X, target = check_X_y(X, y, y_numeric=self.problem_type == "regression", dtype=None)
        if not cols:
            # the additional zeros in the name are because of the variable check in _generate_features,
            # where we check if the column name occurs in the the expression. this would lead to many
            # false positives if we have features x1 and x10...x19 instead of x001...x019.
            cols = [f"x{i:03}" for i in range(X.shape[1])]
        self.original_columns_ = cols
        # transform X into a dataframe (again)
        df = pd.DataFrame(X, columns=cols)
        # possibly convert categorical columns
        df = self._transform_categorical_cols(df)
        # if we're not given specific feateng_cols, then just take all columns except categorical
        if self.feateng_cols:
            fcols = []
            for c in self.feateng_cols:
                if c not in self.original_columns_:
                    raise ValueError(f"[AutoFeat] feateng_col {c} not in df.columns")
                if c in self.categorical_cols_map_:
                    fcols.extend(self.categorical_cols_map_[c])
                else:
                    fcols.append(c)
            self.feateng_cols_ = fcols
        else:
            self.feateng_cols_ = list(df.columns)
        # convert units to proper pint units
        if self.units:
            # need units for only and all feateng columns
            self.units = {c: self.units.get(c, "") for c in self.feateng_cols_}
            # apply pi-theorem -- additional columns are not used for regular feature engineering (for now)!
            df = self._apply_pi_theorem(df)
        # subsample data points and targets in case we'll generate too many features
        # (n_rows * n_cols * 32/8)/1000000000 <= max_gb
        n_cols = n_cols_generated(len(self.feateng_cols_), self.feateng_steps, len(self.transformations))
        n_gb = (len(df) * n_cols) / 250000000
        if self.verbose:
            logging.info(
                f"[AutoFeat] The {self.feateng_steps} step feature engineering process could generate up to {n_cols} features.",
            )
            logging.info(
                f"[AutoFeat] With {len(df)} data points this new feature matrix would use about {n_gb:.2f} gb of space.",
            )
        if self.max_gb and n_gb > self.max_gb:
            n_rows = int(self.max_gb * 250000000 / n_cols)
            if self.verbose:
                logging.info(
                    f"[AutoFeat] As you specified a limit of {self.max_gb:.1f} gb, the number of data points is subsampled to {n_rows}",
                )
            subsample_idx = np.random.permutation(list(df.index))[:n_rows]
            df_subs = df.iloc[subsample_idx]
            df_subs = df_subs.reset_index(drop=True)
            target_sub = target[subsample_idx]
        else:
            df_subs = df.copy()
            target_sub = target.copy()
        # generate features
        df_subs, self.feature_formulas_ = engineer_features(
            df_subs,
            self.feateng_cols_,
            _parse_units(self.units, verbose=self.verbose),  # type: ignore
            self.feateng_steps,
            self.transformations,
            self.verbose,
        )
        # select predictive features
        if self.featsel_runs <= 0:
            if self.verbose:
                logging.warning("[AutoFeat] Not performing feature selection.")
            good_cols = df_subs.columns
        elif self.problem_type in ("regression", "classification"):
            good_cols = select_features(
                df_subs,
                target_sub,
                self.featsel_runs,
                None,
                self.problem_type,
                self.n_jobs,
                self.verbose,
            )
            # if no features were selected, take the original features
            if not good_cols:
                good_cols = list(df.columns)
        else:
            logging.warning(f"[AutoFeat] Unknown problem_type {self.problem_type} - not performing feature selection.")
            good_cols = df_subs.columns
        # filter out those columns that were original features or generated otherwise
        self.new_feat_cols_ = [c for c in good_cols if c not in list(df.columns)]
        self.good_cols_ = good_cols
        # re-generate all good feature again; for all data points this time
        self.feature_functions_ = {}  # type: ignore
        df = self._generate_features(df, self.new_feat_cols_)
        # to prevent an error because sometimes the column names are numpy.str_ instead of normal str
        df.columns = [str(c) for c in df.columns]
        # filter out unnecessary junk from self.feature_formulas_
        self.feature_formulas_ = {f: self.feature_formulas_[f] for f in self.new_feat_cols_ + self.feateng_cols_}
        self.feature_functions_ = {f: self.feature_functions_[f] for f in self.new_feat_cols_}
        self.all_columns_ = list(df.columns)
        # train final prediction model on all selected features
        if self.verbose:
            # final dataframe contains original columns and good additional columns
            logging.info(
                f"[AutoFeat] Final dataframe with {len(df.columns)} feature columns ({len(df.columns) - len(self.original_columns_)} new).",
            )

        # train final prediction model
        if self.problem_type == "regression":
            model = lm.LassoLarsCV(cv=5)
        elif self.problem_type == "classification":
            model = lm.LogisticRegressionCV(cv=5, class_weight="balanced")
        else:
            logging.warning(f"[AutoFeat] Unknown problem_type {self.problem_type} - not fitting a prediction model.")
            model = None
        if model is not None:
            if self.verbose:
                logging.info(f"[AutoFeat] Training final {self.problem_type} model.")
            X = df[self.good_cols_].to_numpy()
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                model.fit(X, target)
            self.prediction_model_ = model
            # sklearn requires a "classes_" attribute
            if self.problem_type == "classification":
                self.classes_ = model.classes_
            if self.verbose:
                # for classification, model.coefs_ is n_classes x n_features, but we need n_features
                coefs = model.coef_ if self.problem_type == "regression" else np.max(np.abs(model.coef_), axis=0)
                weights = dict(zip(self.good_cols_, coefs))
                logging.info("[AutoFeat] Trained model: largest coefficients:")
                logging.info(model.intercept_)
                for c in sorted(weights, key=lambda x: abs(weights[x]), reverse=True):
                    if abs(weights[c]) < 1e-5:
                        break
                    logging.info(f"{weights[c]:.6f} * {c}")
                logging.info(f"[AutoFeat] Final score: {model.score(X, target):.4f}")
        if self.always_return_numpy:
            return df.to_numpy()
        return df

    def fit(self, X: np.ndarray | pd.DataFrame, y: np.ndarray | pd.DataFrame):
        if self.verbose:
            logging.warning("[AutoFeat] This just calls fit_transform() but does not return the transformed dataframe.")
            logging.info("[AutoFeat] It is much more efficient to call fit_transform() instead of fit() and transform()!")
        _ = self.fit_transform(X, y)
        return self

    def transform(self, X: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
        """
        Inputs:
            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)

        Returns:
            - new_df: new pandas dataframe with all the original features (except categorical features transformed
                      into multiple 0/1 columns) and the most promising engineered features. This df can then be
                      used to train your final model.
        """
        check_is_fitted(self, ["feature_formulas_"])
        # store column names as they'll be lost in the other check
        cols = [str(c) for c in X.columns] if isinstance(X, pd.DataFrame) else []
        # check input variables
        X = check_array(X, force_all_finite="allow-nan", dtype=None)
        if not cols:
            cols = [f"x{i:03}" for i in range(X.shape[1])]
        if not cols == self.original_columns_:
            raise ValueError("[AutoFeat] Not the same features as when calling fit.")
        # transform X into a dataframe (again)
        df = pd.DataFrame(X, columns=cols)
        # possibly convert categorical columns
        df = self._transform_categorical_cols(df)
        # possibly apply pi-theorem
        df = self._apply_pi_theorem(df)
        # generate engineered features
        df = self._generate_features(df, self.new_feat_cols_)
        # to prevent an error because sometimes the column names are numpy.str_ instead of normal str
        df.columns = [str(c) for c in df.columns]
        if self.always_return_numpy:
            return df.to_numpy()
        return df

    def predict(self, X: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
        """
        Inputs:
            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)

        Returns:
            - y_pred: predicted targets returned by prediction_model.predict()
        """
        check_is_fitted(self, ["prediction_model_"])
        df = self._X2df(X)
        return self.prediction_model_.predict(df[self.good_cols_].to_numpy())

    def score(self, X: np.ndarray | pd.DataFrame, y: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
        """
        Inputs:
            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)
            - y: pandas dataframe or numpy array with the targets for all n_datapoints
        Returns:
            - R^2/Accuracy returned by prediction_model.score()
        """
        check_is_fitted(self, ["prediction_model_"])
        # store column names as they'll be lost in the other check
        cols = [str(c) for c in X.columns] if isinstance(X, pd.DataFrame) else []
        # check input variables
        X, target = check_X_y(X, y, y_numeric=self.problem_type == "regression", dtype=None)
        if not cols:
            cols = [f"x{i:03}" for i in range(X.shape[1])]
        # transform X into a dataframe (again)
        df = pd.DataFrame(X, columns=cols)
        # do we need to call transform?
        if list(df.columns) != self.all_columns_:
            temp = self.always_return_numpy
            self.always_return_numpy = False
            df = self.transform(df)
            self.always_return_numpy = temp
        return self.prediction_model_.score(df[self.good_cols_].to_numpy(), target)

    @property
    def n_features_in_(self):
        return len(self.feateng_cols_)


class AutoFeatRegressor(AutoFeatModel, BaseEstimator, RegressorMixin):
    """Short-cut initialization for AutoFeatModel with problem_type: regression"""

    def __init__(
        self,
        categorical_cols: list | None = None,
        feateng_cols: list | None = None,
        units: dict | None = None,
        feateng_steps: int = 2,
        featsel_runs: int = 5,
        max_gb: int | None = None,
        transformations: list | tuple = ("1/", "exp", "log", "abs", "sqrt", "^2", "^3"),
        apply_pi_theorem: bool = True,
        always_return_numpy: bool = False,
        n_jobs: int = 1,
        verbose: int = 0,
    ):
        super().__init__(
            "regression",
            categorical_cols,
            feateng_cols,
            units,
            feateng_steps,
            featsel_runs,
            max_gb,
            transformations,
            apply_pi_theorem,
            always_return_numpy,
            n_jobs,
            verbose,
        )


class AutoFeatClassifier(AutoFeatModel, BaseEstimator, ClassifierMixin):
    """Short-cut initialization for AutoFeatModel with problem_type: classification"""

    def __init__(
        self,
        categorical_cols: list | None = None,
        feateng_cols: list | None = None,
        units: dict | None = None,
        feateng_steps: int = 2,
        featsel_runs: int = 5,
        max_gb: int | None = None,
        transformations: list | tuple = ("1/", "exp", "log", "abs", "sqrt", "^2", "^3"),
        apply_pi_theorem: bool = True,
        always_return_numpy: bool = False,
        n_jobs: int = 1,
        verbose: int = 0,
    ):
        super().__init__(
            "classification",
            categorical_cols,
            feateng_cols,
            units,
            feateng_steps,
            featsel_runs,
            max_gb,
            transformations,
            apply_pi_theorem,
            always_return_numpy,
            n_jobs,
            verbose,
        )

    def predict_proba(self, X: np.ndarray | pd.DataFrame) -> np.ndarray | pd.DataFrame:
        """
        Inputs:
            - X: pandas dataframe or numpy array with original features (n_datapoints x n_features)

        Returns:
            - y_pred: predicted targets probabilities returned by prediction_model.predict_proba()
        """
        check_is_fitted(self, ["prediction_model_"])
        df = self._X2df(X)
        return self.prediction_model_.predict_proba(df[self.good_cols_].to_numpy())
